
/*++
 
Module Name:
 
    sample1.c
 
Abstract:
 
    - Display some header information.
    - Build a physical memory dump from a hibernation file.
    - Requires: SandMan library.
 
Environment:
 
    - User mode
 
Revision History:
 
    - Nicolas Ruff, Matthieu Suiche
 
--*/
 
#include <windows.h>
#include <sandman.h>
 
/*++
 
    Function Name: main
 
--*/
 
int main (int argc, TCHAR *argv[])
{
PSANDMAN_OBJECT s;
 
SYSTEMTIME st;
FILETIME ft;
 
ULONG Status;
 
ULONG ulVersion = 0;
ULONG ulMajorVersion = 0;
ULONG ulMinorVersion = 0;
ULONG ulBuild = 0;
 
    printf("    SandMan Project! v%s\n"
           "    (c) %s\n"
           "    %s\n"
           "    Demo: sample1.c\n\n",
            SANDMAN_VERSION, 
            SANDMAN_AUTHOR,
            SANDMAN_WEB);
 
    if (argc != 3)
    {
        printf("Usage: %s hiberfil.sys physicaldump.vmem\n",
            argv[0]);
        return FALSE;
    }

    //
    // Open and parse hiberfil.sys
    //
    s = HiberOpen(argv[1]);
 
    if (s == NULL) {
        printf("Error: Failed to open file.\n");
        return FALSE;
    }
 
    //
    // Display header info
    //
 
    printf("Signature:      %c%c%c%c\n",
        s->FileHdr->Signature[0],
        s->FileHdr->Signature[1],
        s->FileHdr->Signature[2],
        s->FileHdr->Signature[3] );
 
    // Get and format SystemTime
    ft.dwHighDateTime = s->FileHdr->SystemTime.HighPart;
    ft.dwLowDateTime  = s->FileHdr->SystemTime.LowPart;
    FileTimeToSystemTime( &ft, &st );
 
    printf("SystemTime:     %08x%08x [%d/%d/%d (DD/MM/YYYY) %d:%d:%d (UTC)]\n",
        ft.dwHighDateTime, 
        ft.dwLowDateTime,
        st.wDay,
        st.wMonth,
        st.wYear,
        st.wHour,
        st.wMinute,
        st.wSecond );
 
    //
    // Display ProcessorState
    //
 
    printf("\nControl registers flags\n");
 
    printf("CR0: %08x\n",       s->ProcState->SpecialRegisters.u_cr0.Cr0);
    printf("CR0[PAGING]: %d\n", s->ProcState->SpecialRegisters.u_cr0.Paging);
    printf("CR3: %08x\n",       s->ProcState->SpecialRegisters.Cr3);
    printf("CR4: %08x\n",       s->ProcState->SpecialRegisters.u_cr4.Cr4);
    printf("CR4[PSE]: %d\n",    s->ProcState->SpecialRegisters.u_cr4.PageSizeExtensions);
    printf("CR4[PAE]: %d\n",    s->ProcState->SpecialRegisters.u_cr4.PhysicalAddressExtension);
 
 
    ulVersion =  HiberGetVersion(s);
 
    // Get the Windows version.
 
    ulMajorVersion = (DWORD)(LOBYTE(LOWORD(ulVersion)));
    ulMinorVersion = (DWORD)(HIBYTE(LOWORD(ulVersion)));
 
    // Get the build number.
 
    if (ulVersion < 0x80000000)
        ulBuild = (DWORD)(HIWORD(ulVersion));
 
    printf("\nYou want more?\n");
    printf("Windows Version is %d.%d (%d)\n", 
                ulMajorVersion,
                ulMinorVersion,
                ulBuild);
 
    printf("\nPhysical Memory dump.\n");
 
    Status = HiberBuildPhysicalMemoryDump(s, 
        argv[2]);
 
    if (Status == FALSE) return Status;
 
    printf("Memory dump successfuly dumped.\n");
 
    HiberClose(s);
 
    return TRUE;
}