/*++
SandMan framework.
Copyright 2008 (c) Matthieu Suiche. <msuiche[at]gmail.com>

This file is part of SandMan.

SandMan is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

SandMan is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with SandMan.  If not, see <http://www.gnu.org/licenses/>.

Module Name:

    compression.c

Abstract:

    - Windows hibernation functions implementation.
    - BEWARE: Very ugly code. Sorry.

Environment:

    - User mode

Revision History:

    - Matthieu Suiche

--*/

#include "compression.h"

//
// Variable for unimplemented use.
//
#ifndef _XPRESS_DEPENDENCY
#define _XPRESS_DEPENDENCY
#endif

/*++
Function Name: XpressDecode

Overview:

Parameters:
         -

Return Values:
        EAX - Number of decompressed byte (Size2) ou -1

--*/
ULONG
XpressDecode(ULONG DefaultOffset,
             PUCHAR DestinationBuffer, 
             ULONG Size1, 
             ULONG Size2, 
             PUCHAR BlockEntry, 
             ULONG CompressedBlockSize
             )
{

    DECODE_DATA DecodeData;

    ZeroMemory(&DecodeData, sizeof(DECODE_DATA));

    DecodeData.DataOffset = UlongToPtr(DefaultOffset);

    //
    // Little hacks :)
    //
     Size2 += 0x108;

    if ((Size1 == CompressedBlockSize) || (Size1 > 0x10000) || (!Size2))
    {
        return Size2;
    }

    if ((Size1 < CompressedBlockSize) || (!CompressedBlockSize) || (Size1 < 8) || (CompressedBlockSize < 8)) 
    {
        return -1;
    }

    DecodeData.BufferLimit = DestinationBuffer + Size1;

    DecodeData.CompressedBlockSize1 = ULongToPtr(CompressedBlockSize - 1);
    DecodeData.CompressedBlockSize3 = CompressedBlockSize-1;

    DecodeData.DestinationBuffer1 = DestinationBuffer;
    DecodeData.Size2 = ULongToPtr(Size2);
    DecodeData.PageLimit = ULongToPtr(CompressedBlockSize);
    DecodeData.CompressedBlockSize2 = UlongToPtr(CompressedBlockSize - 3);
    DecodeData.DestinationBuffer2 = DestinationBuffer;

    if (Size2 > 0x108)
    {
        DecodeData.DestinationBuffer2 = (DestinationBuffer+Size2) - 0x108;
    }

    DecodeData.AdjustedPadding = UlongToPtr(DefaultOffset);

    if (((CompressedBlockSize + PtrToUlong(BlockEntry))-DefaultOffset) > 0xE8)
    {
        DecodeData.AdjustedPadding = ULongToPtr((CompressedBlockSize + PtrToUlong(BlockEntry)) - 0xE8);
    }

    //
    // Little hacks for uncompression's size
    //
    DecodeData.AdjustedPadding += 0xA8;

    DecodeData.DataOffset = BlockEntry;
    DecodeData.Status = FALSE;
    DecodeData.Status2 = FALSE;

    Decode(&DecodeData);

    if ((DecodeData.PageOffset > DecodeData.PageLimit) || (DecodeData.Status == FALSE) ||
        (DecodeData.Status2 == FALSE) || (DecodeData.DecodedSize > PtrToUlong(DecodeData.Size2)))
    {
        return -1;
    }

    if (DecodeData.Size2 != DecodeData.BufferLimit)
    {
        return Size2;
    }

    return Size2;
}

/*++
Function Name: Encode

Overview:

Parameters:
        -

Return Values:
        -
--*/
ULONG
Encode(PCOMPRESS_CONTEXT CContext)
{
PUCHAR pSourceBuffer, pBuffer, pDestinationBuffer;
ULONG ulOffset, Counter, PreviousValue, PreviousOffset;
PULONG pTable;

ULONG isEqual;
ULONG ulIndex;
ULONG ulSize;

    pTable = (PULONG)CContext;
    pTable +=  sizeof(COMPRESS_CONTEXT)/sizeof(ULONG);
    pSourceBuffer = CContext->Source;

    CContext->u50 = CContext->Source + CContext->u38;

    pSourceBuffer += CContext->u30;
    pDestinationBuffer = CContext->u1C;
    Counter = CContext->Status;

    if (!CContext->u30)
    {
        pDestinationBuffer[0] = pSourceBuffer[0];

        pDestinationBuffer += sizeof(UCHAR);
        pSourceBuffer += sizeof(UCHAR);
        PreviousValue = Counter;
        Counter += Counter;
        if ((Counter < PreviousValue)) 
        {
            ulSize = PtrToUlong(CContext->u20);
            CContext->u20 = (PULONG)pDestinationBuffer;
            pDestinationBuffer += sizeof(ULONG);
            ((PULONG)UlongToPtr(ulSize))[0] = Counter;
            Counter = 1;
        }
    }

    while ((pSourceBuffer < CContext->u50))
    {
        ulIndex = pSourceBuffer[1] + pSourceBuffer[0] * sizeof(ULONG);
        ulIndex = pSourceBuffer[2] + ulIndex * sizeof(USHORT);
        ulSize = PtrToUlong(pSourceBuffer) - DELTA_PAGE;

        PtrToUlong(pBuffer) = pTable[ulIndex];
        pTable[ulIndex] = PtrToUlong(pSourceBuffer);

        if ((PtrToUlong(pBuffer) < ulSize ) ||
            (pBuffer == NULL))
        {
            pDestinationBuffer[0] = pSourceBuffer[0];

            pDestinationBuffer += sizeof(UCHAR);
            pSourceBuffer += sizeof(UCHAR);
            PreviousValue = Counter;
            Counter += Counter;
            if ((Counter < PreviousValue)) 
            {
                ulSize = PtrToUlong(CContext->u20);
                CContext->u20 = (PULONG)pDestinationBuffer;
                pDestinationBuffer += sizeof(ULONG);
                ((PULONG)UlongToPtr(ulSize))[0] = Counter;
                Counter = 1;
            }
            continue;
        }

        ulOffset = ((PULONG)pSourceBuffer)[0];
        ulOffset -= ((PULONG)pBuffer)[0];
        if (ulOffset != 0)
        {
            if (!(ulOffset & 0xFFFFFF)) //test    ulOffset, 0FFFFFFh
            {
                pDestinationBuffer[0] = pSourceBuffer[0];

                pDestinationBuffer += sizeof(UCHAR);
                pSourceBuffer += sizeof(UCHAR);
                PreviousValue = Counter;
                Counter += Counter;
                if ((Counter < PreviousValue)) 
                {
                    ulSize = PtrToUlong(CContext->u20);
                    CContext->u20 = (PULONG)pDestinationBuffer;
                    pDestinationBuffer += sizeof(ULONG);
                    ((PULONG)UlongToPtr(ulSize))[0] = Counter;
                    Counter = 1;
                }
                continue;
            }

            ulSize = 3;
            PtrToUlong(pBuffer) -= PtrToUlong(pSourceBuffer);

            ulOffset = PtrToUlong(pSourceBuffer + ulSize);
            PtrToUlong(pBuffer) =~PtrToUlong(pBuffer);
            pDestinationBuffer += sizeof(USHORT);
            PtrToUlong(pBuffer) <<= 3;
            pSourceBuffer += sizeof(UCHAR);
            pBuffer += ulSize - 3;

            ((PUSHORT)pDestinationBuffer)[-1] = LOWORD(pBuffer);
            PreviousValue = Counter;
            Counter += Counter;
            Counter += 1;  // stc adc

            if (Counter < PreviousValue)
            {
                ulSize = PtrToUlong(CContext->u20);
                CContext->u20 = (PULONG)pDestinationBuffer;
                pDestinationBuffer += sizeof(ULONG);
                ((PULONG)UlongToPtr(ulSize))[0] = Counter;
                Counter = 1;
            }
        }

        while (TRUE)
        {
            if (ulOffset > PtrToUlong(CContext->u4C))
            {
                PreviousOffset = ulOffset;

                while (pSourceBuffer < CContext->u4C)
                {
                    ulIndex = pSourceBuffer[1] + (pSourceBuffer[0] * sizeof(ULONG));
                    ulIndex = pSourceBuffer[2] + ulIndex * sizeof(USHORT);

                    pTable[ulIndex] = PtrToUlong(pSourceBuffer);

                    pSourceBuffer += sizeof(UCHAR);
                }
                PtrToUlong(pSourceBuffer) = PreviousOffset;

                break;  //jmp _start
            }

            if (ulOffset != 0)
            {
                do
                {
                    ulIndex = pSourceBuffer[1] + (pSourceBuffer[0] * sizeof(ULONG));
                    ulIndex = pSourceBuffer[2] + ulIndex * sizeof(USHORT);

                    pTable[ulIndex] = PtrToUlong(pSourceBuffer);

                    pSourceBuffer += sizeof(UCHAR);
                } while (PtrToUlong(pSourceBuffer) != ulOffset);

                break;   // jmp start
            }
        // ---------------------------------------------------------------------------

            ulOffset = ((PULONG)pSourceBuffer)[1];
            ulOffset -= ((PULONG)pBuffer)[1];
            if (ulOffset != 0)
            {

                ulSize = _bsf(ulSize, ulOffset);

                PtrToUlong(pBuffer) -= PtrToUlong(pSourceBuffer);
                ulSize >>= 3;

                PtrToUlong(pBuffer) = ~PtrToUlong(pBuffer);

                ulSize += sizeof(ULONG);
                pDestinationBuffer += sizeof(USHORT);

                ulOffset = PtrToUlong(pSourceBuffer + ulSize);

                PtrToUlong(pBuffer) <<= 3;
                pSourceBuffer += sizeof(UCHAR);

                PtrToUlong(pBuffer) += ulSize - 3;
                ((PUSHORT)pDestinationBuffer)[-1] = LOWORD(pBuffer);

                PreviousValue = Counter;
                Counter += Counter;
                Counter += 1; //stc adc
                if (!(Counter < PreviousValue)) continue; // jnb

                ulSize = PtrToUlong(CContext->u20);
                CContext->u20 = (PULONG)pDestinationBuffer;

                pDestinationBuffer += sizeof(ULONG);

                ((PULONG)UlongToPtr(ulSize))[0] = Counter;
                Counter = 1;

                continue;
            }
        // ---------------------------------------------------------------------------

            ulOffset = PtrToUlong(pSourceBuffer);

            pSourceBuffer += 2 * sizeof(ULONG);
            pBuffer += 2 * sizeof(ULONG);

            ulSize = PtrToUlong(CContext->SourceLimit - pSourceBuffer);

            isEqual = 0;
            _cmpsb(pSourceBuffer, pBuffer, ulSize, isEqual)
            if (isEqual)
            {
                pSourceBuffer += sizeof(UCHAR);
                pBuffer += sizeof(UCHAR);
            }

            ulSize = PtrToUlong(pSourceBuffer - 1);
            PtrToUlong(pBuffer) -= PtrToUlong(pSourceBuffer);
            ulSize -= ulOffset;
            PtrToUlong(pSourceBuffer) = ulOffset;

            if (ulSize < 0xA)
            {
                ulOffset = PtrToUlong(pSourceBuffer + ulSize);
                PtrToUlong(pBuffer) =~PtrToUlong(pBuffer);
                pDestinationBuffer += sizeof(USHORT);
                PtrToUlong(pBuffer) <<= 3;
                pSourceBuffer += sizeof(UCHAR);
                pBuffer += ulSize - 3;

                ((PUSHORT)pDestinationBuffer)[-1] = LOWORD(pBuffer);
                PreviousValue = Counter;
                Counter += Counter;
                Counter += 1;  // stc adc

                if (Counter < PreviousValue)
                {
                    ulSize = PtrToUlong(CContext->u20);
                    CContext->u20 = (PULONG)pDestinationBuffer;
                    pDestinationBuffer += sizeof(ULONG);
                    ((PULONG)UlongToPtr(ulSize))[0] = Counter;
                    Counter = 1;
                }

                continue;
            }

            PtrToUlong(pBuffer) = ~PtrToUlong(pBuffer);
            ulOffset = PtrToUlong(pSourceBuffer + ulSize);

            PtrToUlong(pBuffer) <<= 3;
            ulSize -= 0xA;

            pBuffer += 7;

            PreviousOffset = ulOffset;
            ((PUSHORT)pDestinationBuffer)[0] = LOWORD(pBuffer);

            (BYTE)ulOffset = LOBYTE(ulSize); // al, cl

            if (ulSize > 0xF) ulOffset = 0xF;

            pBuffer = UlongToPtr(CContext->u00);
            pDestinationBuffer += sizeof(USHORT);
            if (pBuffer != NULL)
            {
                ulOffset = LOBYTE(ulOffset) << 4;
                CContext->u00 = 0;
                pBuffer[0] |= LOBYTE(ulOffset);
            }

            CContext->u00 = PtrToUlong(pDestinationBuffer);
            pDestinationBuffer[0] = LOBYTE(ulOffset);
            pDestinationBuffer += sizeof(UCHAR);

            ulSize -= 0xF;

            if (((signed)ulSize > 0))
            {
                pDestinationBuffer[0] = LOBYTE(ulSize);    
                pDestinationBuffer += sizeof(UCHAR);
                if (ulSize > 0xFF)
                {
                    ulSize += 0x16;
                    pDestinationBuffer[-1] = 0xFF;
                    ((PUSHORT)pDestinationBuffer)[0] = LOWORD(ulSize);
                    pDestinationBuffer += sizeof(USHORT);
                }
            }

            pSourceBuffer += sizeof(UCHAR);
            ulOffset = PreviousOffset;
            PreviousValue = Counter;
            Counter += Counter;
            Counter += 1;
            if (!(Counter < PreviousValue)) continue;

            ulSize = PtrToUlong(CContext->u20);
            CContext->u20 = (PULONG)pDestinationBuffer;
            pDestinationBuffer += sizeof(ULONG);
            ((PULONG)UlongToPtr(ulSize))[0] = Counter;
            Counter = 1;
        }
}

    CContext->u1C = pDestinationBuffer;
    CContext->Status = Counter;
    PtrToUlong(pSourceBuffer) -= PtrToUlong(CContext->Source);
    PtrToUlong(CContext->u30) = PtrToUlong(pSourceBuffer);

return CContext->Status;
}

/*++
Function Name: Decode

Overview:

Parameters:
        -

Return Values:
        -
--*/
ULONG Decode(PDECODE_DATA Block
              )
{
ULONG Pos;
PUCHAR DestinationBuffer2;
PUCHAR Const;
PUCHAR Size2;
PUCHAR AdjustedPadding;
PUCHAR PageLimit;
PUCHAR CompressedBlockSize1;
PUCHAR CompressedBlockSize2;
PDECODE_DATA DecodeData;

ULONG ulStatus, ulCounter, ulMonkey, ulPreviousStatus;
PUCHAR pSourceBuffer, pBuffer, pDestinationBuffer, ulNoodles;


    DecodeData = Block;

    Size2 = DecodeData->Size2;
    DestinationBuffer2 = DecodeData->DestinationBuffer2;
    AdjustedPadding = DecodeData->AdjustedPadding;
    PageLimit = DecodeData->PageLimit;
    CompressedBlockSize1 = DecodeData->CompressedBlockSize1;
    CompressedBlockSize2 = DecodeData->CompressedBlockSize2;
    ulMonkey = 0;
    Pos = 0;
    Const = DecodeData->DestinationBuffer1;
    pDestinationBuffer = DecodeData->DestinationBuffer1;
    pSourceBuffer = DecodeData->DataOffset;

    ulStatus = 0;

    while (TRUE)
    {

        if (!(pSourceBuffer < AdjustedPadding) || !(pDestinationBuffer < DestinationBuffer2))
        {
            if (!(pSourceBuffer < CompressedBlockSize2))
            {
                DecodeData->Status = 0;
                return DecodeData->Status;
            }

            ulStatus = ((PULONG)pSourceBuffer)[0];
            pSourceBuffer += sizeof(ULONG);
            ulPreviousStatus = ulStatus;
            ulStatus += ulStatus + 1;

            if ((signed)ulPreviousStatus > 0) goto _11;

            goto _12;
        }


        ulStatus = ((PULONG)pSourceBuffer)[0];
        pSourceBuffer += sizeof(ULONG);
        ulPreviousStatus = ulStatus;
        ulStatus = ulStatus * 2 + 1; // Counter
        if (((signed)ulPreviousStatus >= 0))
        {

            do
            {
                //pDestinationBuffer[-1] = LOBYTE(ulCounter);

                _2:                                     
                pDestinationBuffer += sizeof(UCHAR);
                (UCHAR)ulCounter = pSourceBuffer[0];
                pSourceBuffer += sizeof(UCHAR);
                ulPreviousStatus = ulStatus;
                ulStatus += ulStatus;

                pDestinationBuffer[-1] = LOBYTE(ulCounter);
            } while (ulStatus > ulPreviousStatus);         // jnb

            //pDestinationBuffer[-1] = LOBYTE(ulCounter);

            if (ulStatus == 0) continue; // goto _start;
        }

        _3:
        ulMonkey = 0;
        (USHORT)ulMonkey = ((PUSHORT)pSourceBuffer)[0];
        ulCounter = ulMonkey;
        ulMonkey >>= 3;
        pSourceBuffer += sizeof(USHORT);
        ulMonkey = ~ulMonkey;
        ulCounter &= 7;

        pBuffer = pDestinationBuffer + ulMonkey;
        if (LOBYTE(ulCounter) > 5) goto _5;
        if (pBuffer < Const)
        {
            DecodeData->Status = 0;
            return DecodeData->Status;
        }

        if (!(ulMonkey < -3))
        {
            ulMonkey = ((PULONG)pBuffer)[0];

            ulCounter += 3;
            _movsb(pDestinationBuffer, pBuffer, ulCounter);

            ulPreviousStatus = ulStatus;
            ulStatus += ulStatus;
            if (ulStatus > ulPreviousStatus) goto _2;
            if(ulStatus != 0) goto _3;
            continue;
        }

        ((PULONG)pDestinationBuffer)[0] = ((PULONG)pBuffer)[0];

        ((PULONG)pDestinationBuffer)[1] = ((PULONG)pBuffer)[1];

        pDestinationBuffer += ulCounter + 3;
        ulPreviousStatus = ulStatus;
        ulStatus += ulStatus;

        if (ulStatus > ulPreviousStatus) goto _2;
        if(ulStatus != 0) goto _3;

        continue;
        // ------------------------------------------------------------------------

        // ------------------------------------------------------------------------

        _5:

        if ( pBuffer < Const) 
        {
            DecodeData->Status = 0;
            return DecodeData->Status;
        }

        ulMonkey = Pos;
        
        if (LOBYTE(ulCounter) != 7) goto _8;
        if (ulMonkey == 0)
        {
            ulCounter = 0;
            (UCHAR)ulCounter = pSourceBuffer[0];
            ulMonkey = PtrToUlong(pSourceBuffer);
            ulCounter &= 0xF;
            pSourceBuffer += sizeof(UCHAR);
        }
        else
        {
            ulCounter = 0;
            ulNoodles = UlongToPtr(ulMonkey);
            (UCHAR)ulCounter = (UCHAR)ulNoodles[0];
            ulMonkey= 0;
            ulCounter >>= 4;
        }

        Pos = ulMonkey;
        ulCounter += 7;

        if (LOBYTE(ulCounter - 0x7) == 0xF) goto _9;


        _8:
        ulMonkey = PtrToUlong(pDestinationBuffer) + ulCounter + 3;
        ulCounter += 3;

        if (!(ulMonkey < PtrToUlong(DestinationBuffer2))) goto _16;
        _movsb(pDestinationBuffer, pBuffer, ulCounter);

        ulPreviousStatus = ulStatus;
        ulStatus += ulStatus;

        if (ulStatus > ulPreviousStatus) goto _2;
        if(ulStatus != 0) goto _3;

        continue; // goto _start;

        // ------------------------------------------------------------------------

        _9:                                     
        ulCounter = 0;
        (UCHAR)ulCounter = pSourceBuffer[0];
        pSourceBuffer += sizeof(BYTE);
        ulCounter += 0x16;

        if ((LOBYTE(ulCounter - 0x16)) != 0xFF) goto _8;

        //ulCounter += 0x16;
        ulCounter = 0;
        ulCounter = ((PUSHORT)pSourceBuffer)[0];
        pSourceBuffer += sizeof(USHORT);

        if (!(ulCounter < 0x115)) goto _8;

        //-------------------------------------------------------------------------
        do
        {
            //pDestinationBuffer[-1] = LOBYTE(ulCounter);

            _11:                                    
            if (!(pDestinationBuffer < Size2))
            {
                ulStatus = 1;
                DecodeData->PageOffset = pSourceBuffer;
                DecodeData->DecodedSize = PtrToUlong(pDestinationBuffer);
                DecodeData->Status = ulStatus;

                return DecodeData->Status;
            }

            if (!(pSourceBuffer < PageLimit))
            {
                DecodeData->Status = 0;
                return DecodeData->Status;
            }

            pDestinationBuffer += sizeof(UCHAR);
            (UCHAR)ulCounter = pSourceBuffer[0];
            pSourceBuffer += sizeof(UCHAR);
            ulPreviousStatus = ulStatus;
            ulStatus += ulStatus;

            pDestinationBuffer[-1] = LOBYTE(ulCounter);

            if (ulStatus > ulPreviousStatus) continue;

            //pDestinationBuffer[-1] = LOBYTE(ulCounter);

            if (LOBYTE(ulCounter) == 0)
            {
                if (!(pSourceBuffer < CompressedBlockSize2))
                {
                    DecodeData->Status = 0;
                    return DecodeData->Status;
                }

                ulStatus = ((PULONG)pSourceBuffer)[0];
                pSourceBuffer += sizeof(ULONG);
                ulPreviousStatus = ulStatus;
                ulStatus += ulStatus + 1;
            }
        } while ((signed)ulPreviousStatus > 0);

        _12:
        if (!(pDestinationBuffer < Size2))
        {
            ulStatus = 1;
            if (pDestinationBuffer == DecodeData->BufferLimit)
            {
                DecodeData->Status2 = ulStatus;
            }

            DecodeData->PageOffset = pSourceBuffer;
            DecodeData->DecodedSize = PtrToUlong(pDestinationBuffer);
            return DecodeData->DecodedSize;
        }

        if (!(pSourceBuffer < CompressedBlockSize1))
        {
            DecodeData->Status = 0;
            return DecodeData->Status;
        }
        
        ulMonkey = 0;
        ulMonkey = ((PUSHORT)pSourceBuffer)[0];
        ulCounter = ulMonkey;
        ulCounter >>= 3;

        pSourceBuffer += sizeof(USHORT);
        ulMonkey = ~ulMonkey;
        ulCounter &= 0x7;
        pBuffer = pDestinationBuffer + ulMonkey;

        if (pBuffer < Const)
        {
            DecodeData->Status = 0;
            return DecodeData->Status;
        }

        ulMonkey = Pos;

        if (LOBYTE(ulCounter) == 7)
        {
            if (ulMonkey == 0)
            {
                if (!(pSourceBuffer < PageLimit))
                {
                    DecodeData->Status = 0;
                    return DecodeData->Status;
                }

                ulCounter = 0;
                (UCHAR)ulCounter = pSourceBuffer[0];
                ulMonkey = PtrToUlong(pSourceBuffer);
                ulCounter &= 0xF;
                pSourceBuffer += sizeof(UCHAR);
            }
            else
            {
                ulCounter = 0;
                ulNoodles = UlongToPtr(ulMonkey);
                (UCHAR)ulCounter = ulNoodles[0];
                ulMonkey = 0;
                ulCounter >>= 4;
            }
            Pos = ulMonkey;
            ulCounter += 7;
            if ((LOBYTE(ulCounter - 0x7) ) == 0xF) 
            {

                if ((pSourceBuffer < PageLimit))
                {
                    ulCounter = 0;
                    (UCHAR)ulCounter = pSourceBuffer[0];
                    pSourceBuffer += sizeof(UCHAR);
                    ulCounter += 0x16;
                    if ((LOBYTE(ulCounter - 0x16)) == 0xFF)
                    {
                        if ((pSourceBuffer < CompressedBlockSize1))
                        {
                            ulCounter = 0;
                            ulCounter = ((PUSHORT)pSourceBuffer)[0];
                            pSourceBuffer += sizeof(USHORT);
                            if ((ulCounter < 0x115))
                            {
                                DecodeData->Status = 0;
                                return DecodeData->Status;
                            }
                        }
                    }
                }
                else
                {
                    DecodeData->Status = 0;
                    return DecodeData->Status;
                }
            }
        }

        ulMonkey = PtrToUlong(pDestinationBuffer + ulCounter + 3);
        ulCounter += 3;

        _16:                                    
        if (ulMonkey <= PtrToUlong(Size2))
        {
            _movsb(pDestinationBuffer, pBuffer, ulCounter);

            ulPreviousStatus = ulStatus;
            ulStatus += ulStatus;
            if (ulStatus > ulPreviousStatus) goto _11;
            if (ulStatus != 0) goto _12;
        }
        else
        {
            ulCounter = PtrToUlong(Size2);
            ulCounter -= PtrToUlong(pDestinationBuffer);
            // pDestinationBuffer > ulCounter because of userland address space
            ulCounter &= 0x0000FFFF;
            _movsb(pDestinationBuffer, pBuffer, ulCounter);

            ulStatus = 1;
            DecodeData->PageOffset = pSourceBuffer;
            DecodeData->DecodedSize = PtrToUlong(pDestinationBuffer);
            DecodeData->Status = ulStatus;
            return DecodeData->Status;
        }
    }

    //
    // Never reached.
    //
    DecodeData->Status = ulStatus;
    return ulStatus;
}


/*++
Function Name: XpressEncodeCreate

Overview:

Parameters:
        -

Return Values:
        - COMPRESSION_WORKSPACE
--*/
PHIBER_CONTEXT 
XpressEncodeCreate(ULONG Size,
                   ULONG Arg2,
                   PVOID pPopAllocateHiberContextCallback,
                   ULONG CompressionRelated // Compression Type
                   )
{
PHIBER_CONTEXT HiberContext;
ULONG         eax, ecx, ebx, esi;

    if (Size > 0x10000) return FALSE;

    ebx = CompressionRelated;
    ecx = 0;
    eax = DEFAULT_CONTEXT_SIZE;

    if (ebx > 9)
    {
        ebx = 9;
        eax = 0x10000;
    }
    else if ((ebx >= 2) && (ebx < 9))
    {
        eax = Size*2;
        ecx = 0x10000;
    }

    HiberContext = VirtualAlloc (NULL,
                                (eax+((ecx+0x194)&0xFFFFFFE0)+0x38),
                                MEM_COMMIT | MEM_RESERVE,
                                PAGE_READWRITE);

    if (!HiberContext) return FALSE;

    HiberContext->Magic = XPRESS_ENCODE_MAGIC;
    //
    // Not Implemented. We use VirtualAlloc instead of it.
    //
    HiberContext->pPopAllocateHiberContextCallback = NULL;
    HiberContext->CompressionType = ebx;
    HiberContext->SelfHiberContext = HiberContext;
    eax = (PtrToUlong(HiberContext));
    eax += (0x20-((PtrToUlong(&HiberContext->u018) & 0x1F)));
    esi = ((ecx+0x194)&0xFFFFFFE0);

    HiberContext->CContextLimit = UlongToPtr(esi + eax);
    HiberContext->Buffer2 = UlongToPtr(eax);

    return HiberContext;
}

/*++
Function Name: XpressEncode

Overview:

Parameters:
            -

Return Values:
            -

--*/
ULONG
XpressEncode(PHIBER_CONTEXT HiberContext,
             LPVOID Destination,
             ULONG Size1,
             PUCHAR Source,
             ULONG Size2,
             LPVOID _IoResume,
             ULONG Arg7,
             ULONG Arg8
             )
{
ULONG eax;
ULONG Status;
PCOMPRESS_CONTEXT CContext;


    Status = FALSE;

    if (!HiberContext) return Status;

    //
    // Check if magic code match or not
    //
    if (HiberContext->Magic != XPRESS_ENCODE_MAGIC) return Status;

    if ((Size2 <= 8)                    || 
        (Size1 < 8)                     ||
        (!Destination)                  ||
        (!Source)) return Size2;

    CContext = (PCOMPRESS_CONTEXT)(HiberContext->CContextLimit - sizeof(COMPRESS_CONTEXT));
    RtlZeroMemory(CContext, sizeof(COMPRESS_CONTEXT));

    CContext->Source = Source;

    CContext->SourceLimit = Source + Size2;

    CContext->u4C = CContext->SourceLimit + ~2;
    CContext->u10 = Destination;
    CContext->u0C = Destination;
    CContext->u38 = Size2;
    CContext->u34 = Size2;
    CContext->u48 = (CContext->SourceLimit - 0x10);
    CContext->u14 = HiberContext->Buffer2;  // Destination
    CContext->u1C = HiberContext->Buffer2;
    CContext->CompressionType = HiberContext->CompressionType;
    CContext->u30 = 0;
    CContext->u18 = Destination;
    
    Destination = Encode;

    RtlZeroMemory(HiberContext->CContextLimit, DEFAULT_CONTEXT_SIZE);

    CContext->u18 = CContext->u10;
    CContext->u20 = (PULONG)CContext->u10;
    CContext->u1C = CContext->u10 + sizeof(ULONG);
    eax = PtrToUlong(CContext->u0C) 
        - PtrToUlong(CContext->u1C)
        + Size1 - 8
        - ((PtrToUlong(CContext->u0C) - PtrToUlong(CContext->u1C) + Size1 - 8 + 7) >> 3);

    CContext->Status = TRUE;
    if ( eax > 16)
    {

        while (TRUE)
        {

            if (CContext->u30 > CContext->u34) break;

            eax += CContext->u30;

            if (eax > (CContext->u34 + ~7))
            {
                CContext->u38 = (CContext->u34 + ~7);

                if (CContext->u30 > CContext->u38) break;
            }

            CompressBuffer(Destination,
                           0,           //_IoResume
                           Arg7,        // 0
                           CContext);

            eax = PtrToUlong(CContext->u0C) 
                - PtrToUlong(CContext->u1C)
                + Size1 
                - 8
                + ((PtrToUlong(CContext->u0C) - PtrToUlong(CContext->u1C) + Size1 - 8 + 7) >> 3);

            if (eax <= 16) return Size2;
        }

        if (CContext->u30 < CContext->u34)
        {
            do
            {
                *CContext->u1C = *(CContext->Source + CContext->u30);
                CContext->u1C++;

                eax = 2 * CContext->Status;

                if ( CContext->Status < 0 )
                {
                    *CContext->u20 = eax;
                    CContext->u20 = (PULONG)CContext->u1C;
                    CContext->u20++;
                    eax = 1;
                }

                CContext->u30++;
                CContext->Status = eax;

            } while(CContext->u30 > CContext->u34);
        }

        EndCompression(CContext);

        eax = PtrToUlong(CContext->u1C - CContext->u18);

        if (eax < Size1)
        {

            CContext->u10 += eax;

            Status = PtrToUlong(CContext->u10 - CContext->u0C);

        }
    }

    return Status;
}

/*
Function Name: CompressBuffer

Overview:

Parameters:
            -

Return Values:
            -

--*/
VOID
CompressBuffer(LPVOID Callbackfunction,
               LPVOID _IoResume,
               ULONG Unused,
               PCOMPRESS_CONTEXT CContext // Additionnal argument
               )
{
VOID (WINAPI *Compressing)(PCOMPRESS_CONTEXT);

    Compressing = Callbackfunction;

    //
    // We don't catch IoResume() anymore.
    //
    Compressing(CContext);

}

/*
Function Name: EndCompression

Overview:

Parameters:
            -

Return Values:
            -
*/
VOID 
EndCompression(PCOMPRESS_CONTEXT CContext
               )
{
ULONG Status, Counter;
PULONG Limit, Odd;

    Status = CContext->Status;
    Counter = CContext->Status;
    Odd = (PULONG)CContext->u1C;
    Limit = Odd;

    while (Limit == Odd)
    {
        Status = (2 * Counter )|1;

        if ((signed)Counter < 0)
        {
            *CContext->u20 = Status;
            CContext->u20 = Limit;
            Limit += 4;
            Status = TRUE;
        }
        Counter = Status;
    }

    CContext->Status = Status;

}

/*
Function Name: GetXpressBlockSize

Overview:
            - This functions aims to read the Size dword, and convert it
              into an human readable size format.

Parameters:
            - PIMAGE_XPRESS_HEADER pIXH

Return Values:
            -
*/
ULONG
GetXpressBlockSize(PIMAGE_XPRESS_HEADER pIXH
                   )
{
ULONG Size;

    Size = (pIXH->Size.u0B << 24)
         + (pIXH->Size.u0A << 16)
         + (pIXH->Size.u09 << 8);
    Size >>= 10;
    Size += 1;


    if ((Size % 8) == 0)
        return Size;
    return (Size & ~7) + 8;
}
